"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          15-05-2018
//      Created for Project :   FLAME
"""

from json import loads
from re import compile, IGNORECASE

CONFIG_ATTRIBUTES = ('aggregator_report_period', 'aggregator_database_name', 'aggregator_database_url')  # all of the configuration attributes - to be used as dictionary keys

RUNNING_FLAG = 'aggregator_running'  # Attribute for storing the flag, which shows whether the aggregator is running or not - to be used as a dictionary key

PROCESS_ATTRIBUTE = 'aggregator_process'  # Attribute for storing the process object of the aggregator - to be used as a dictionary key

# a 'malformed' running state of the aggregator is when the configuration is updated, but the aggregator is not restarted so it is running with an old version of the conf.
MALFORMED_FLAG = 'malformed'  # Attribute for storing the flag, which shows whether the aggregator is running in an malformed state or not - to be used as a dictionary key

# used to indicate a malformed configuration message
COMMENT_ATTRIBUTE = 'comment'
COMMENT_VALUE = 'Aggregator is running in a malformed state - it uses an old version of the configuration. Please, restart it so that the updated configuration is used.'

# the attributes of the JSON response body that are expected when querying round trip time
ROUND_TRIP_ATTRIBUTES = ('media_service', 'start_timestamp', 'end_timestamp')


URL_REGEX = compile(
    r'^https?://'  # http:// or https://
    r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain, e.g. example.domain.com
    r'localhost|'  # or localhost...
    r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # or IP address (IPv4 format)
    r'(?::\d{2,5})?'  # optional port number
    r'(?:[/?#][^\s]*)?$',  # URL path or query parameters
    IGNORECASE)


def validate_config_content(configuration):
    """
    A utility function to validate a configuration string representing a JSON dictionary.

    :param configuration: the configuration string to validate
    :return the validated configuration dictionary object with the values converted to their required type
    :raise AssertionError: if the argument is not a valid configuration
    """

    global CONFIG_ATTRIBUTES

    try:
        configuration = loads(configuration)
    except:
        raise AssertionError("Configuration must be a JSON object.")

    assert len(configuration) == len(CONFIG_ATTRIBUTES), "Configuration mustn't contain more attributes than the required ones."

    for attribute in CONFIG_ATTRIBUTES:
        assert attribute in configuration, "Required attribute not found in the request content."

    assert type(configuration.get('aggregator_report_period')) == int, "Report period must be an integer, received {0} instead.".format(configuration.get('aggregator_report_period'))

    assert configuration.get('aggregator_report_period') > 0, "Report period must be a positive integer, received {0} instead.".format(configuration.get('aggregator_report_period'))

    assert URL_REGEX.match(configuration.get('aggregator_database_url')) is not None, "The aggregator must have a valid database URL in its configuration, received {0} instead.".format(configuration.get('aggregator_database_url'))

    return configuration


def validate_action_content(content):
    """
    A utility function to validate a content string representing a JSON dictionary.

    :param content: the content string to validate
    :return: the validated content dictionary
    :raise AssertionError: if the argument is not a valid json content
    """

    try:
        content = loads(content)
    except:
        raise AssertionError("Content must be a JSON object.")

    assert len(content) == 1, "Content mustn't contain more attributes than the required one."

    assert content['action'] in ('start', 'stop', 'restart')

    return content


def validate_round_trip_query_params(params):
    """
    A utility function to validate a dictionary of parameters.

    :param params: the params dict to validate
    :return: the validated parameters dictionary
    :raise AssertionError: if the argument is not a valid json content
   """

    global ROUND_TRIP_ATTRIBUTES

    assert len(params) == len(ROUND_TRIP_ATTRIBUTES), "Content mustn't contain more attributes than the required ones."

    for attribute in ROUND_TRIP_ATTRIBUTES:
        assert attribute in params, "Required attribute not found in the request content."

    return params


def generate_e2e_delay_report(path_id, source_sfr, target_sfr, endpoint, sf_instance, delay_forward, delay_reverse, delay_service, avg_request_size, avg_response_size, avg_bandwidth, time):
    """
    Generates a combined averaged measurement about the e2e delay and its contributing parts

    :param path_id: The path identifier, which is a bidirectional path ID for the request and the response path
    :param source_sfr: source service router
    :param target_sfr: target service router
    :param endpoint: endpoint of the media component
    :param sf_instance: service function instance (media component)
    :param delay_forward: Path delay (Forward direction)
    :param delay_reverse: Path delay (Reverse direction)
    :param delay_service: the media service component response time
    :param avg_request_size: averaged request size
    :param avg_response_size: averaged response size
    :param avg_bandwidth: averaged bandwidth
    :param time: measurement timestamp
    :return: a list of dict-formatted reports to post on influx
    """

    result = [{"measurement": "e2e_delays",
               "tags": {
                   "path_ID": path_id,
                   "source_SFR": source_sfr,
                   "target_SFR": target_sfr,
                   "endpoint": endpoint,
                   "sf_instance": sf_instance
               },
               "fields": {
                   "delay_forward": float(delay_forward),
                   "delay_reverse": float(delay_reverse),
                   "delay_service": float(delay_service),
                   "avg_request_size": float(avg_request_size),
                   "avg_response_size": float(avg_response_size),
                   "avg_bandwidth": float(avg_bandwidth)
               },
               "time": int(1000000000*time)
               }]

    return result
